from contextlib import suppress
from typing import Any, AsyncGenerator, Optional

import pytest

from aiocouch import CouchDB, Database, Document, NotFoundError


@pytest.fixture
async def couchdb_user_account(couchdb: CouchDB) -> AsyncGenerator[CouchDB, None]:
    users = await couchdb.create("_users", exists_ok=True)

    doc = await users.create("org.couchdb.user:aiocouch_test_user")
    doc["name"] = "aiocouch_test_user"
    doc["password"] = "aiocouch_test_user"
    doc["roles"] = ["aiocouch_test_role"]
    doc["type"] = "user"

    await doc.save()
    assert doc["_id"] == "org.couchdb.user:aiocouch_test_user"
    assert doc.id == "org.couchdb.user:aiocouch_test_user"

    yield couchdb

    await doc.delete()


@pytest.fixture
async def couchdb_utf8_user_account(couchdb: CouchDB) -> AsyncGenerator[CouchDB, None]:
    users = await couchdb.create("_users", exists_ok=True)

    doc = await users.create("org.couchdb.user:aiocouch_test_user_🛋️")
    doc["name"] = "aiocouch_test_user_🛋️"
    doc["password"] = "aiocouch_test_user_🛋️"
    doc["roles"] = ["aiocouch_test_role"]
    doc["type"] = "user"

    await doc.save()
    assert doc["_id"] == "org.couchdb.user:aiocouch_test_user_🛋️"
    assert doc.id == "org.couchdb.user:aiocouch_test_user_🛋️"

    yield couchdb

    await doc.delete()


@pytest.fixture
async def couchdb() -> AsyncGenerator[CouchDB, None]:
    import asyncio
    import os

    from aiocouch import CouchDB

    try:
        hostname = os.environ["COUCHDB_HOST"]
    except KeyError:
        hostname = "http://localhost:5984"

    user: Optional[str] = None

    with suppress(KeyError):
        user = os.environ["COUCHDB_USER"]

    try:
        password = os.environ["COUCHDB_PASS"]
    except KeyError:
        password = ""

    async with CouchDB(
        hostname, user=user, password=password, loop=asyncio.get_event_loop()
    ) as couchdb:
        yield couchdb

    # Give the couchdb server some time, so the assumptions of an empty database isn't
    # broken for subsequent tests. This delay used to be part of the disconnect
    # procedure, but #6 removed that. Therefore, we do sleep for the previous amount
    # here. This should give the server some time to delete the old fixtures.
    await asyncio.sleep(0.250)


@pytest.fixture
async def couchdb_with_user_access(
    couchdb_user_account: CouchDB,
) -> AsyncGenerator[CouchDB, None]:
    import os

    from aiocouch import CouchDB

    try:
        hostname = os.environ["COUCHDB_HOST"]
    except KeyError:
        hostname = "http://localhost:5984"

    async with CouchDB(
        hostname, user="aiocouch_test_user", password="aiocouch_test_user"
    ) as couchdb:
        yield couchdb


@pytest.fixture
def database_id() -> str:
    return "aiocouch_test_fixture_database"


@pytest.fixture
async def database(
    couchdb: CouchDB, database_id: str
) -> AsyncGenerator[Database, None]:
    database = await couchdb.create(database_id)

    yield database

    await database.delete()


@pytest.fixture
async def filled_database(database: Database) -> AsyncGenerator[Database, None]:
    doc = await database.create("foo")
    doc["bar"] = True
    doc["bar2"] = 3
    await doc.save()

    doc = await database.create("foo2")
    doc["bar"] = True
    doc["bar2"] = 1
    await doc.save()

    doc = await database.create("baz")
    doc["bar"] = False
    doc["bar2"] = 4
    await doc.save()

    doc = await database.create("baz2")
    doc["bar"] = True
    doc["bar2"] = 2
    await doc.save()

    yield database


@pytest.fixture
async def filled_database_with_view(
    filled_database: Database,
) -> AsyncGenerator[Database, None]:
    ddoc = await filled_database.design_doc("test_ddoc")
    await ddoc.create_view(
        "null_view",
        map_function="function (doc) { emit(doc._id, null); }",
    )
    await ddoc.create_view(
        "full_view",
        map_function="function (doc) { emit(doc._id, doc); }",
    )
    await ddoc.create_view(
        "bar_view",
        map_function="function (doc) { emit(doc._id, doc.bar); }",
    )

    yield filled_database


@pytest.fixture
async def large_filled_database(database: Database) -> AsyncGenerator[Database, None]:
    async with database.update_docs(
        [f"doc{i}" for i in range(2000)], create=True
    ) as docs:
        async for doc in docs:
            doc["llama"] = "awesome"

    yield database


@pytest.fixture
async def doc(database: Database) -> AsyncGenerator[Document, None]:
    doc = await database.create("foo")
    yield doc



# Copyright (c) 2019, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools
from contextlib import suppress
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

import aiohttp
from typing_extensions import Protocol


class Endpoint(Protocol):
    @property
    def id(self) -> str:  # pragma: no cover
        ...

    @property
    def endpoint(self) -> str:  # pragma: no cover
        ...

    @property
    def _data(self) -> Dict[str, Any]:  # pragma: no cover
        ...

    @property
    def _document(self) -> "Endpoint":  # pragma: no cover
        ...


class BadRequestError(ValueError):
    """Represents a 400 HTTP status code returned from the server"""

    pass


class UnauthorizedError(PermissionError):
    """Represents a 401 HTTP status code returned from the server"""

    pass


class ForbiddenError(PermissionError):
    """Represents a 403 HTTP status code returned from the server"""

    pass


class NotFoundError(KeyError):
    """Represents a 404 HTTP status code returned from the server"""

    pass


class ConflictError(KeyError):
    """Represents a 409 HTTP status code returned from the server"""

    pass


class PreconditionFailedError(ValueError):
    """Represents a 412 HTTP status code returned from the server"""

    pass


class UnsupportedMediaTypeError(ValueError):
    """Represents a 415 HTTP status code returned from the server"""

    pass


class ExpectationFailedError(ValueError):
    """Represents a 417 HTTP status code returned from the server"""

    pass


class RemoteResponseError(aiohttp.ClientResponseError):
    def __init__(self, reason: Union[str, None], *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.reason = reason


def raise_for_endpoint(
    endpoint: Endpoint,
    message: str,
    exception: RemoteResponseError,
    exception_type: Optional[Type[Exception]] = None,
) -> NoReturn:
    if exception_type is None:
        if exception.status == 400:
            exception_type = BadRequestError
        elif exception.status == 401:
            exception_type = UnauthorizedError
        elif exception.status == 403:
            exception_type = ForbiddenError
        elif exception.status == 404:
            exception_type = NotFoundError
        elif exception.status == 409:
            exception_type = ConflictError
        elif exception.status == 412:
            exception_type = PreconditionFailedError
        elif exception.status == 415:
            exception_type = UnsupportedMediaTypeError
        elif exception.status == 417:
            exception_type = ExpectationFailedError
        else:
            raise ValueError(
                "Something went wrong, but I couldn't deduce the type of exception nor "
                "format a nice exception for you. I'm sorry."
            ) from exception

    message_input = {}

    with suppress(AttributeError):
        message_input["reason"] = exception.reason
    message_input["reason"] = message_input.get("reason", exception.message)
    with suppress(AttributeError):
        message_input["id"] = endpoint.id
        message_input["endpoint"] = endpoint.endpoint
        message_input["rev"] = cast(str, endpoint._data.get("_rev"))
    with suppress(AttributeError):
        message_input["document_id"] = endpoint._document.id
        message_input["document_rev"] = cast(str, endpoint._document._data.get("_rev"))

    raise exception_type(message.format(**message_input)) from exception


FuncT = TypeVar("FuncT", bound=Callable[..., Any])


def raises(
    status: int, message: str, exception_type: Optional[Type[Exception]] = None
) -> Callable[[FuncT], FuncT]:
    def decorator_raises(func: FuncT) -> FuncT:
        @functools.wraps(func)
        async def wrapper(endpoint: Endpoint, *args: Any, **kwargs: Any) -> Any:
            try:
                return await func(endpoint, *args, **kwargs)
            except RemoteResponseError as exception:
                if status == exception.status:
                    raise_for_endpoint(endpoint, message, exception, exception_type)
                raise exception

        return cast(FuncT, wrapper)

    return decorator_raises


def generator_raises(
    status: int, message: str, exception_type: Optional[Type[Exception]] = None
) -> Callable[[FuncT], FuncT]:
    def decorator_raises(func: FuncT) -> FuncT:
        @functools.wraps(func)
        async def wrapper(
            endpoint: Endpoint, *args: Any, **kwargs: Any
        ) -> AsyncGenerator[Any, None]:
            try:
                async for data in func(endpoint, *args, **kwargs):
                    yield data
            except RemoteResponseError as exception:
                if status == exception.status:
                    raise_for_endpoint(endpoint, message, exception, exception_type)
                raise exception

        return cast(FuncT, wrapper)

    return decorator_raises



import pytest

from aiocouch import Database

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_null_view_keys(filled_database_with_view: Database) -> None:
    values = [
        key
        async for key in filled_database_with_view.view(
            "test_ddoc", "null_view"
        ).akeys()
    ]

    assert len(values) == 4

    assert values[0] == "baz"
    assert values[1] == "baz2"
    assert values[2] == "foo"
    assert values[3] == "foo2"


async def test_null_view_ids(filled_database_with_view: Database) -> None:
    values = [
        key
        async for key in filled_database_with_view.view("test_ddoc", "null_view").ids()
    ]

    assert len(values) == 4

    assert values[0] == "baz"
    assert values[1] == "baz2"
    assert values[2] == "foo"
    assert values[3] == "foo2"


async def test_null_view_mixed_ids(filled_database_with_view: Database) -> None:
    values = [
        key
        async for key in filled_database_with_view.view("test_ddoc", "null_view").ids(
            ids=["baz", "baz2", "foo", "not_existing"]
        )
    ]

    assert len(values) == 3

    assert values[0] == "baz"
    assert values[1] == "baz2"
    assert values[2] == "foo"


async def test_null_view_docs(filled_database_with_view: Database) -> None:
    values = [
        doc.id
        async for doc in filled_database_with_view.view("test_ddoc", "null_view").docs()
    ]

    assert len(values) == 4

    assert values[0] == "baz"
    assert values[1] == "baz2"
    assert values[2] == "foo"
    assert values[3] == "foo2"


async def test_null_view_docs_with_deleted(filled_database_with_view: Database) -> None:
    doc = await filled_database_with_view["baz"]
    await doc.delete()

    docs = [
        doc
        async for doc in filled_database_with_view.view("test_ddoc", "null_view").docs(
            ids=["baz"]
        )
    ]

    assert len(docs) == 0


async def test_create_existing_view(filled_database_with_view: Database) -> None:
    ddoc = await filled_database_with_view.design_doc("test_ddoc", exists_ok=True)

    with pytest.raises(KeyError):
        await ddoc.create_view(
            "null_view",
            map_function="function (doc) { emit(doc._id, null); }",
        )


async def test_create_view_with_reduce(database: Database) -> None:
    ddoc = await database.design_doc("my_test_ddoc")

    await ddoc.create_view(
        "my_test_view",
        map_function="function (doc) { emit(doc._id, null); }",
        reduce_function="_count",
    )


async def test_view_avalues(filled_database_with_view: Database) -> None:
    docs = [
        doc
        async for doc in filled_database_with_view.view(
            "test_ddoc", "null_view"
        ).avalues()
    ]

    assert len(docs) == 4


async def test_holding_docs_wrong(filled_database_with_view: Database) -> None:
    with pytest.raises(ValueError):
        async for doc in filled_database_with_view.view("test_ddoc", "null_view").docs(
            ids=["baz"], prefix="foo"
        ):
            pass


async def test_null_view_ids_with_prefix(filled_database_with_view: Database) -> None:
    values = [
        key
        async for key in filled_database_with_view.view("test_ddoc", "null_view").ids(
            prefix="ba"
        )
    ]

    assert len(values) == 2

    assert values[0] == "baz"
    assert values[1] == "baz2"


async def test_view_response_contains_update_seq(
    filled_database_with_view: Database,
) -> None:
    response = await filled_database_with_view.view("test_ddoc", "null_view").get(
        update_seq=True
    )

    assert response.update_seq is not None


async def test_view_keys_param_rejected(
    filled_database_with_view: Database,
) -> None:
    view = filled_database_with_view.view("test_ddoc", "null_view")

    with pytest.raises(AttributeError):
        async for doc in view.docs(key='"baz"'):
            pass

    with pytest.raises(AttributeError):
        async for id in view.ids(key='"baz"'):
            pass

    with pytest.raises(AttributeError):
        async for doc in view.docs(keys='["baz"]'):
            pass

    with pytest.raises(AttributeError):
        async for id in view.ids(keys='["baz"]'):
            pass


async def test_view_get_with_keys(
    filled_database_with_view: Database,
) -> None:
    # Note: This directly uses the get method of the View to provide the key
    # and keys request query parameter. You likely don't want to do this. Use
    # the ids() or docs() method with the ids parameter instead.

    view = filled_database_with_view.view("test_ddoc", "null_view")
    response = await view.get(keys='["baz"]')

    keys = [id for id in response.keys()]

    assert len(keys) == 1
    assert keys[0] == "baz"


async def test_view_get_with_key(
    filled_database_with_view: Database,
) -> None:
    # Note: This directly uses the get method of the View to provide the key
    # and keys request query parameter. You likely don't want to do this. Use
    # the ids() or docs() method with the ids parameter instead.

    view = filled_database_with_view.view("test_ddoc", "null_view")
    response = await view.get(key='"baz"')

    keys = [id for id in response.keys()]

    assert len(keys) == 1
    assert keys[0] == "baz"



from typing import cast

import pytest

from aiocouch.database import Database
from aiocouch.document import SecurityDocument

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_create_document(database: Database) -> None:
    doc = await database.create("test_document")

    # doc not yet saved, so it shouldn't be listed in the database
    assert doc.id not in [key async for key in database.akeys()]

    await doc.save()

    # doc was saved, so it should be listed in the database
    assert doc.id in [key async for key in database.akeys()]


async def test_create_document_with_data(database: Database) -> None:
    doc = await database.create("test_document", data={"foo": 1234})

    assert doc["foo"] == 1234

    # doc not yet saved, so it shouldn't be listed in the database
    assert doc.id not in [key async for key in database.akeys()]

    await doc.save()

    # doc was saved, so it should be listed in the database
    assert doc.id in [key async for key in database.akeys()]


async def test_getitem_for_existing(filled_database: Database) -> None:
    doc = await filled_database["foo"]

    assert doc["bar"] is True
    assert doc["_id"] == doc.id == "foo"
    assert doc["_rev"].startswith("1-")
    assert doc._dirty_cache is False


async def test_getitem_for_non_existing(database: Database) -> None:
    with pytest.raises(KeyError):
        await database["foo"]


async def test_create_for_existing(filled_database: Database) -> None:
    with pytest.raises(KeyError):
        await filled_database.create("foo")


async def test_create_for_existing_exists_true(filled_database: Database) -> None:
    doc = await filled_database.create("foo", exists_ok=True)

    assert doc["bar"] is True
    assert doc["_id"] == doc.id == "foo"
    assert doc["_rev"].startswith("1-")
    assert doc._dirty_cache is False


async def test_create_for_non_existing_exists_true(database: Database) -> None:
    doc = await database.create("foo", exists_ok=True)

    assert doc["_id"] == doc.id == "foo"
    assert "_rev" not in doc
    assert doc._dirty_cache is True


async def test_create_for_existing_exists_true_data(filled_database: Database) -> None:
    doc = await filled_database.create(
        "foo", exists_ok=True, data={"bar": False, "fox": "red"}
    )

    assert doc["bar"] is True
    assert "fox" not in doc
    assert doc["_id"] == doc.id == "foo"
    assert doc["_rev"].startswith("1-")
    assert doc._dirty_cache is False


async def test_create_for_non_existing_exists_true_data(database: Database) -> None:
    doc = await database.create("foo", exists_ok=True, data={"foo": 1234})

    assert doc["_id"] == doc.id == "foo"
    assert doc["foo"] == 1234
    assert "_rev" not in doc
    assert doc._dirty_cache is True


async def test_get_for_existing(filled_database: Database) -> None:
    doc = await filled_database.get("foo")

    assert doc["bar"] is True
    assert doc["_id"] == doc.id == "foo"
    assert doc["_rev"].startswith("1-")
    assert doc._dirty_cache is False


async def test_get_for_non_existing(database: Database) -> None:
    with pytest.raises(KeyError):
        await database.get("foo")


async def test_get_with_bad_rev(filled_database: Database) -> None:
    from aiocouch.exception import BadRequestError

    with pytest.raises(BadRequestError):
        await filled_database.get("foo", rev="dgsahfgesogfesgfh")


async def test_get_for_non_existing_with_empty_default(database: Database) -> None:
    doc = await database.get("foo", default={})

    assert doc["_id"] == doc.id == "foo"
    assert "_rev" not in doc
    assert doc._dirty_cache is True


async def test_get_for_non_existing_with_default(database: Database) -> None:
    doc = await database.get("foo", default={"jumbo": "dumbo", "value": 42})

    assert doc["_id"] == doc.id == "foo"
    assert "_rev" not in doc
    assert doc["jumbo"] == "dumbo"
    assert doc["value"] == 42
    assert doc._dirty_cache is True


async def test_akeys_with_prefix(filled_database: Database) -> None:
    keys = [key async for key in filled_database.akeys(prefix="ba")]

    assert len(keys) == 2
    assert (sorted(keys)) == ["baz", "baz2"]


async def test_akeys_with_ids(filled_database: Database) -> None:
    keys = [key async for key in filled_database.akeys(ids=["foo", "baz", "halloween"])]

    assert len(keys) == 2
    assert keys == ["foo", "baz"]


async def test_saved_docs_in_filled_db(filled_database: Database) -> None:
    keys = [key async for key in filled_database.akeys()]

    assert len(keys) == 4
    assert sorted(keys) == ["baz", "baz2", "foo", "foo2"]


async def test_docs_on_empty(database: Database) -> None:
    all_docs = [doc async for doc in database.docs([])]

    assert all_docs == []


async def test_docs_filtered(filled_database: Database) -> None:
    keys = [doc.id async for doc in filled_database.docs(["foo", "baz"])]

    assert len(keys) == 2
    assert "foo" in keys
    assert "baz" in keys


async def test_docs_on_non_existent(database: Database) -> None:
    docs = [doc async for doc in database.docs(["foo"], create=True)]

    assert len(docs) == 1
    doc = docs[0]
    assert doc._dirty_cache is True
    assert doc.id == "foo"


async def test_docs_with_prefix(filled_database: Database) -> None:
    keys = [doc.id async for doc in filled_database.docs(prefix="ba")]

    assert len(keys) == 2
    assert (sorted(keys)) == ["baz", "baz2"]


async def test_docs_on_deleted(filled_database: Database) -> None:
    doc = await filled_database["foo"]
    await doc.delete()

    with pytest.raises(KeyError):
        async for doc in filled_database.docs(["foo"]):
            assert False

    async for doc in filled_database.docs(["foo"], create=True):
        assert doc.id == "foo"
        assert doc.exists is False


async def test_docs_with_no_ids(filled_database: Database) -> None:
    keys = [doc.id async for doc in filled_database.docs()]

    assert len(keys) == 4
    assert sorted(keys) == ["baz", "baz2", "foo", "foo2"]


async def test_docs_do_not_contain_ddocs(filled_database_with_view: Database) -> None:
    keys = [doc.id async for doc in filled_database_with_view.docs()]

    assert "_design/test_ddoc" not in keys


async def test_docs_contain_ddocs_with_param(
    filled_database_with_view: Database,
) -> None:
    keys = [doc.id async for doc in filled_database_with_view.docs(include_ddocs=True)]

    assert "_design/test_ddoc" in keys


async def test_find(filled_database: Database) -> None:
    matching_docs = [doc async for doc in filled_database.find({"bar": True})]

    assert len(matching_docs) == 3

    from aiocouch.document import Document

    matching_keys = []
    for doc in matching_docs:
        assert isinstance(doc, Document)
        matching_keys.append(doc.id)

    assert "foo" in matching_keys
    assert "foo2" in matching_keys
    assert "baz2" in matching_keys


async def test_find_limited(filled_database: Database) -> None:
    matching_docs = [doc async for doc in filled_database.find({"bar": True}, limit=1)]

    assert len(matching_docs) == 1

    from aiocouch.document import Document

    matching_keys = []
    for doc in matching_docs:
        assert isinstance(doc, Document)
        matching_keys.append(doc.id)

    assert "baz2" in matching_keys


async def test_purge(filled_database: Database) -> None:
    from aiocouch.exception import NotFoundError

    doc = await filled_database["foo"]
    await filled_database._purge({doc.id: [doc.rev]})

    with pytest.raises(NotFoundError):
        await doc.info()


async def test_find_fields_parameter_gets_rejected(database: Database) -> None:
    with pytest.raises(ValueError):
        [doc async for doc in database.find({"bar": True}, fields="anything")]


async def test_find_sorted(filled_database: Database) -> None:
    await filled_database.index({"fields": ["bar2"]})

    matching_docs = [
        doc async for doc in filled_database.find({"bar": True}, sort=["bar2"])
    ]

    assert len(matching_docs) == 3
    assert matching_docs[0]["bar2"] == 1
    assert matching_docs[1]["bar2"] == 2
    assert matching_docs[2]["bar2"] == 3


async def test_alldocs_values(filled_database: Database) -> None:
    values = [key async for key, value in filled_database.all_docs.aitems()]

    assert len(values) == 4

    assert values[0] == "baz"
    assert values[1] == "baz2"
    assert values[2] == "foo"
    assert values[3] == "foo2"


async def test_values_for_filled(filled_database: Database) -> None:
    keys = [doc.id async for doc in filled_database.values()]

    assert len(keys) == 4
    assert sorted(keys) == ["baz", "baz2", "foo", "foo2"]


async def test_values_for_filled_limited(filled_database: Database) -> None:
    keys = [doc.id async for doc in filled_database.values(limit=1)]

    assert len(keys) == 1
    assert keys == ["baz"]


async def test_many_docs(large_filled_database: Database) -> None:
    keys = [key async for key in large_filled_database.akeys()]
    assert len(keys) == 2000

    docs = [doc async for doc in large_filled_database.docs()]
    assert len(docs) == 2000

    find_docs = [
        doc async for doc in large_filled_database.find(selector={"llama": "awesome"})
    ]
    assert len(find_docs) == 2000


async def test_get_design_doc(filled_database_with_view: Database) -> None:
    await filled_database_with_view.design_doc("test_ddoc2")
    await filled_database_with_view.design_doc("test_ddoc", exists_ok=True)

    with pytest.raises(KeyError):
        await filled_database_with_view.design_doc("test_ddoc")


async def test_set_invalid_design_doc_key(filled_database_with_view: Database) -> None:
    ddoc = await filled_database_with_view.design_doc("test_ddoc", exists_ok=True)

    with pytest.raises(KeyError):
        ddoc["foo"] = "bar"


async def test_get_security(database: Database) -> None:
    sec = await database.security()
    sec2 = SecurityDocument(cast(Database, None), data=await database._get_security())

    assert sec.members is None
    assert sec2.members is None
    assert sec.admins is None
    assert sec2.admins is None

    assert sec.member_roles is None or sec.member_roles == ["_admin"]
    assert sec2.member_roles is None or sec2.member_roles == ["_admin"]
    assert sec.admin_roles is None or sec.admin_roles == ["_admin"]
    assert sec2.admin_roles is None or sec2.admin_roles == ["_admin"]


async def test_set_security(database: Database) -> None:
    sec = await database.security()

    result = await database._put_security(sec._data)

    assert "ok" in result
    assert result["ok"] is True


async def test_security_add_members(database: Database) -> None:
    sec = await database.security()
    sec.add_member("foobert")

    await sec.save()

    sec2 = await database.security()
    assert sec2.members is not None
    assert "foobert" in sec2.members

    sec2.remove_member("foobert")
    await sec2.save()

    sec3 = await database.security()
    assert sec3.members is not None
    assert "foobert" not in sec3.members


async def test_security_remove_member(database: Database) -> None:
    sec = await database.security()
    with pytest.raises(KeyError):
        sec.remove_member("foobert")


async def test_security_add_admins(database: Database) -> None:
    sec = await database.security()
    sec.add_admin("foobert")

    await sec.save()

    sec2 = await database.security()
    assert sec2.admins is not None
    assert "foobert" in sec2.admins

    sec2.remove_admin("foobert")
    await sec2.save()

    sec3 = await database.security()
    assert sec3.admins is not None
    assert "foobert" not in sec3.admins


async def test_security_remove_admin(database: Database) -> None:
    sec = await database.security()
    with pytest.raises(KeyError):
        sec.remove_admin("foobert")


async def test_security_stays_empty(database: Database) -> None:
    sec = await database.security()
    old_data = sec._data
    await sec.save()

    sec2 = await database.security()
    assert sec2._data == old_data



# Copyright (c) 2020, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Any, AsyncGenerator, Dict, Optional, cast

from . import database
from .document import Document


class FindRequestChunk:
    def __init__(
        self,
        database: "database.Database",
        *,
        data: Dict[str, Any],
        pagination_size: int,
    ):
        self.database = database
        self.data = data
        self.pagination_size = pagination_size

    @property
    async def docs(self) -> AsyncGenerator[Document, None]:
        for res in self.data["docs"]:
            doc = Document(self.database, res["_id"])
            doc._update_cache(res)
            yield doc

    @property
    def bookmark(self) -> str:
        return cast(str, self.data["bookmark"])

    @property
    def is_last_chunk(self) -> bool:
        return len(self.data["docs"]) < self.pagination_size


class FindRequest:
    def __init__(
        self,
        database: "database.Database",
        selector: Any,
        *,
        limit: Optional[int] = None,
        **params: Any,
    ):
        self.database = database
        self.selector = selector
        self.limit = limit
        self.params = params

    async def __aiter__(self) -> AsyncGenerator[Document, None]:
        pagination_size = self.limit if self.limit is not None else 10000

        while True:
            chunk = FindRequestChunk(
                self.database,
                data=await self.database._find(
                    self.selector,
                    limit=pagination_size,
                    **self.params,
                ),
                pagination_size=pagination_size,
            )

            self.params["bookmark"] = chunk.bookmark

            async for doc in chunk.docs:
                yield doc

            if self.limit is not None or chunk.is_last_chunk:
                break



# Copyright (c) 2019, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import asyncio
import json
from contextlib import suppress
from typing import Any, AsyncGenerator, Dict, List, Optional, Tuple, Union, cast
from urllib.parse import quote

import aiohttp

from . import database, document
from .exception import NotFoundError, RemoteResponseError, generator_raises, raises
from .typing import JsonDict


class HTTPResponse:
    """Represents an HTTP response from the CouchDB server."""

    status: int
    """The HTTP response status, usually 200, 201 or 202"""

    headers: Dict[str, str]
    """The HTTP headers of the response"""

    def __init__(self, resp: aiohttp.client.ClientResponse):
        self.status = resp.status
        self.headers = dict(resp.headers)

    @property
    def etag(self) -> Optional[str]:
        """Convenient property to access the ETag header in a usable format"""
        return self.headers["Etag"][1:-1] if "Etag" in self.headers else None


RequestResult = Tuple[HTTPResponse, Union[bytes, JsonDict]]


def _quote_id(id: str) -> str:
    return quote(id, safe="")


def _stringify_params(params: Optional[JsonDict]) -> Optional[JsonDict]:
    if params is None:
        return None
    result = {}
    for key, value in params.items():
        if value is None:
            continue
        elif value is True:
            value = "true"
        elif value is False:
            value = "false"

        result[key] = value

    return result


class RemoteServer:
    def __init__(
        self,
        server: str,
        *,
        user: Optional[str] = None,
        password: Optional[str] = None,
        cookie: Optional[str] = None,
        **kwargs: Any,
    ):
        self._server = server
        auth = aiohttp.BasicAuth(user, password, "utf-8") if user and password else None
        headers = {"Cookie": "AuthSession=" + cookie} if cookie else None
        self._http_session = aiohttp.ClientSession(headers=headers, auth=auth, **kwargs)

    async def _get(
        self, path: str, *, params: Optional[JsonDict] = None
    ) -> RequestResult:
        return await self._request("GET", path, params=params)

    async def _get_bytes(
        self, path: str, *, params: Optional[JsonDict] = None
    ) -> RequestResult:
        return await self._request("GET", path, params=params, return_json=False)

    async def _put(
        self,
        path: str,
        *,
        data: Optional[JsonDict] = None,
        params: Optional[JsonDict] = None,
    ) -> RequestResult:
        return await self._request("PUT", path, json=data, params=params)

    async def _put_bytes(
        self,
        path: str,
        *,
        data: bytes,
        content_type: str,
        params: Optional[JsonDict] = None,
    ) -> RequestResult:
        return await self._request(
            "PUT",
            path,
            data=data,
            params=params,
            headers={"Content-Type": content_type},
        )

    async def _post(
        self, path: str, *, data: JsonDict, params: Optional[JsonDict] = None
    ) -> RequestResult:
        return await self._request("POST", path, json=data, params=params)

    async def _delete(
        self, path: str, *, params: Optional[JsonDict] = None
    ) -> RequestResult:
        return await self._request("DELETE", path, params=params)

    async def _head(
        self,
        path: str,
        *,
        params: Optional[JsonDict] = None,
        **kwargs: Any,
    ) -> RequestResult:
        return await self._request("HEAD", path, params=params, **kwargs)

    async def _request(
        self,
        method: str,
        path: str,
        *,
        params: Optional[JsonDict] = None,
        return_json: bool = True,
        **kwargs: Any,
    ) -> RequestResult:
        kwargs["params"] = _stringify_params(params) if params else {}

        async with self._http_session.request(
            method, url=f"{self._server}{path}", **kwargs
        ) as resp:
            reason = None
            with suppress(Exception):
                reason = (await resp.json())["reason"]
            try:
                resp.raise_for_status()
            except aiohttp.ClientResponseError:
                raise RemoteResponseError(
                    reason,
                    resp.request_info,
                    resp.history,
                    status=resp.status,
                    message=resp.reason,
                    headers=resp.headers,
                )

            return (
                HTTPResponse(resp),
                await resp.json() if return_json else await resp.read(),
            )

    async def _streamed_request(
        self,
        method: str,
        path: str,
        *,
        params: Optional[JsonDict] = None,
        **kwargs: Any,
    ) -> AsyncGenerator[JsonDict, None]:
        kwargs["params"] = _stringify_params(params) if params else {}
        kwargs.setdefault("timeout", aiohttp.ClientTimeout())

        async with self._http_session.request(
            method, url=f"{self._server}{path}", **kwargs
        ) as resp:
            try:
                resp.raise_for_status()
            except aiohttp.ClientResponseError:
                raise RemoteResponseError(
                    None,
                    resp.request_info,
                    resp.history,
                    status=resp.status,
                    message=resp.reason,
                    headers=resp.headers,
                )

            async for line in resp.content:
                # this should only happen for empty lines
                with suppress(json.JSONDecodeError):
                    yield json.loads(line)

    @raises(401, "Invalid credentials")
    @raises(403, "Access forbidden: {reason}")
    async def _all_dbs(self, **params: Any) -> List[str]:
        _, json = await self._get("/_all_dbs", params=params)
        assert not isinstance(json, bytes)
        return cast(List[str], json)

    async def close(self) -> None:
        # If ClientSession has TLS/SSL connections, it is needed to wait 250 ms
        # before closing, see https://github.com/aio-libs/aiohttp/issues/1925.
        has_ssl_conn = self._http_session.connector and any(
            any(hasattr(handler.transport, "_ssl_protocol") for handler, _ in conn)
            for conn in self._http_session.connector._conns.values()
        )
        await self._http_session.close()
        await asyncio.sleep(0.250 if has_ssl_conn else 0)

    @raises(401, "Invalid credentials")
    @raises(403, "Access forbidden: {reason}")
    async def _info(self) -> JsonDict:
        _, json = await self._get("/")
        assert not isinstance(json, bytes)
        return json

    @raises(401, "Authentication failed, check provided credentials.")
    @raises(403, "Access forbidden: {reason}")
    async def _check_session(self) -> RequestResult:
        return await self._get("/_session")


class RemoteDatabase:
    def __init__(self, remote: RemoteServer, id: str):
        self.id = id
        self._remote = remote

    @property
    def endpoint(self) -> str:
        return f"/{_quote_id(self.id)}"

    @raises(401, "Invalid credentials")
    @raises(403, "Access forbidden: {reason}")
    async def _exists(self) -> bool:
        try:
            await self._remote._head(self.endpoint)
            return True
        except RemoteResponseError as e:
            if e.status == 404:
                return False
            else:
                raise e

    @raises(401, "Invalid credentials")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Requested database not found ({id})")
    async def _get(self) -> JsonDict:
        _, json = await self._remote._get(self.endpoint)
        assert not isinstance(json, bytes)
        return json

    @raises(400, "Invalid database name")
    @raises(401, "CouchDB Server Administrator privileges required")
    @raises(403, "Access forbidden: {reason}")
    @raises(412, "Database already exists")
    async def _put(self, **params: Any) -> JsonDict:
        _, json = await self._remote._put(self.endpoint, params=params)
        assert not isinstance(json, bytes)
        return json

    @raises(400, "Invalid database name or forgotten document id by accident")
    @raises(401, "CouchDB Server Administrator privileges required")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Database doesn't exist or invalid database name ({id})")
    async def _delete(self) -> None:
        await self._remote._delete(self.endpoint)

    @raises(400, "The request provided invalid JSON data or invalid query parameter")
    @raises(401, "Read permission required")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Invalid database name")
    @raises(415, "Bad Content-Type value")
    async def _bulk_get(self, docs: List[JsonDict], **params: Any) -> JsonDict:
        _, json = await self._remote._post(
            f"{self.endpoint}/_bulk_get",
            data={"docs": docs},
            params=params,
        )
        assert not isinstance(json, bytes)
        return json

    @raises(400, "The request provided invalid JSON data")
    @raises(401, "Invalid credentials")
    @raises(403, "Access forbidden: {reason}")
    @raises(417, "At least one document was rejected by the validation function")
    async def _bulk_docs(self, docs: List[JsonDict], **data: Any) -> JsonDict:
        data["docs"] = docs
        _, json = await self._remote._post(
            f"{self.endpoint}/_bulk_docs",
            data=data,
        )
        assert not isinstance(json, bytes)
        return json

    @raises(400, "Invalid request")
    @raises(401, "Read privilege required for document '{id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(500, "Query execution failed", RuntimeError)
    async def _find(self, selector: Any, **data: Any) -> JsonDict:
        data["selector"] = selector
        _, json = await self._remote._post(
            f"{self.endpoint}/_find",
            data=data,
        )
        assert not isinstance(json, bytes)
        return json

    @raises(400, "Invalid request")
    @raises(401, "Admin permission required")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Database not found")
    @raises(500, "Execution error")
    async def _index(self, index: JsonDict, **data: Any) -> JsonDict:
        data["index"] = index
        _, json = await self._remote._post(
            f"{self.endpoint}/_index",
            data=data,
        )
        assert not isinstance(json, bytes)
        return json

    @raises(401, "Invalid credentials")
    @raises(403, "Access forbidden: {reason}")
    async def _get_security(self) -> JsonDict:
        _, json = await self._remote._get(f"{self.endpoint}/_security")
        assert not isinstance(json, bytes)
        return json

    @raises(401, "Invalid credentials")
    @raises(403, "Access forbidden: {reason}")
    async def _put_security(self, doc: JsonDict) -> JsonDict:
        _, json = await self._remote._put(
            f"{self.endpoint}/_security",
            data=doc,
        )
        assert not isinstance(json, bytes)
        return json

    @generator_raises(400, "Invalid request")
    @generator_raises(403, "Access forbidden: {reason}")
    async def _changes(self, **params: Any) -> AsyncGenerator[JsonDict, None]:
        if "feed" in params and params["feed"] == "continuous":
            params.setdefault("heartbeat", True)
            async for data in self._remote._streamed_request(
                "GET", f"{self.endpoint}/_changes", params=params
            ):
                yield data
        else:
            _, json = await self._remote._get(
                f"{self.endpoint}/_changes", params=params
            )
            assert not isinstance(json, bytes)
            for result in json["results"]:
                yield result

    @raises(400, "Invalid database or JSON payload")
    @raises(403, "Access forbidden: {reason}")
    @raises(415, "Bad Content-Type header value")
    @raises(500, "Internal server error or timeout")
    async def _purge(self, docs: JsonDict, **params: Any) -> JsonDict:
        _, json = await self._remote._post(
            f"{self.endpoint}/_purge", data=docs, params=params
        )
        assert not isinstance(json, bytes)
        return json


class RemoteDocument:
    def __init__(self, database: "database.Database", id: str):
        self._database = database
        self.id = id
        self._data: Optional[JsonDict] = None

    @property
    def endpoint(self) -> str:
        return f"{self._database.endpoint}/{_quote_id(self.id)}"

    @raises(401, "Read privilege required for document '{id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Document {id} was not found")
    async def _head(self) -> None:
        await self._database._remote._head(self.endpoint)

    @raises(401, "Read privilege required for document '{id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Document {id} was not found")
    async def _info(self) -> JsonDict:
        response, _ = await self._database._remote._head(self.endpoint)
        assert self._data is not None
        return {
            "ok": True,
            "id": self._data["_id"],
            "rev": response.headers["Etag"][1:-1],
        }

    async def _exists(self) -> bool:
        try:
            await self._head()
            return True
        except NotFoundError:
            return False

    @raises(400, "The format of the request or revision was invalid")
    @raises(401, "Read privilege required for document '{id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Document {id} was not found")
    async def _get(self, **params: Any) -> JsonDict:
        _, json = await self._database._remote._get(
            self.endpoint,
            params=params,
        )
        assert not isinstance(json, bytes)
        return json

    @raises(400, "The format of the request or revision was invalid")
    @raises(401, "Write privilege required for document '{id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Specified database or document ID doesn't exists ({endpoint})")
    @raises(
        409,
        "Document with the specified ID ({id}) already exists or specified revision "
        "{rev} is not latest for target document",
    )
    async def _put(
        self, data: JsonDict, **params: Any
    ) -> Tuple[HTTPResponse, JsonDict]:
        response, json = await self._database._remote._put(
            self.endpoint,
            data=data,
            params=params,
        )
        assert not isinstance(json, bytes)
        return (response, json)

    @raises(400, "Invalid request body or parameters")
    @raises(401, "Write privilege required for document '{id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Specified database or document ID doesn't exists ({endpoint})")
    @raises(
        409, "Specified revision ({rev}) is not the latest for target document '{id}'"
    )
    async def _delete(self, rev: str, **params: Any) -> Tuple[HTTPResponse, JsonDict]:
        params["rev"] = rev
        response, json = await self._database._remote._delete(
            self.endpoint,
            params=params,
        )
        assert not isinstance(json, bytes)
        return (response, json)

    @raises(400, "Invalid request body or parameters")
    @raises(401, "Read or write privileges required")
    @raises(403, "Access forbidden: {reason}")
    @raises(
        404, "Specified database, document ID or revision doesn't exists ({endpoint})"
    )
    @raises(
        409,
        "Document with the specified ID already exists or specified revision is not "
        "latest for target document",
    )
    async def _copy(
        self, destination: str, **params: Any
    ) -> Tuple[HTTPResponse, JsonDict]:
        response, json = await self._database._remote._request(
            "COPY",
            self.endpoint,
            params=params,
            headers={"Destination": destination},
        )
        assert not isinstance(json, bytes)
        return (response, json)

    @raises(400, "The format of the request or revision was invalid")
    @raises(401, "Read privilege required for document '{id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Document {id} was not found")
    async def _conflicts(self) -> List[str]:
        json = await self._get(conflicts=True)
        assert not isinstance(json, bytes)
        return cast(List[str], json.get("_conflicts", []))

    @raises(400, "The format of the request or revision was invalid")
    @raises(401, "Read privilege required for document '{id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Document {id} was not found")
    async def _revs(self) -> List[str]:
        json = await self._get(revs_info=True)
        assert not isinstance(json, bytes)
        return cast(List[str], [item["rev"] for item in json.get("_revs_info", [])])


class RemoteAttachment:
    def __init__(self, document: "document.Document", id: str):
        self._document = document
        self.id = id
        self.content_type: Optional[str] = None

    @property
    def endpoint(self) -> str:
        return f"{self._document.endpoint}/{_quote_id(self.id)}"

    @raises(401, "Read privilege required for document '{document_id}'")
    @raises(403, "Access forbidden: {reason}")
    async def _exists(self) -> bool:
        try:
            response, _ = await self._document._database._remote._head(
                self.endpoint, return_json=False
            )
            self.content_type = response.headers["Content-Type"]
            return True
        except RemoteResponseError as e:
            if e.status == 404:
                return False
            else:
                raise e

    @raises(400, "Invalid request parameters")
    @raises(401, "Read privilege required for document '{document_id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Document '{document_id}' or attachment '{id}' doesn't exists")
    async def _get(self, **params: Any) -> bytes:
        response, data = await self._document._database._remote._get_bytes(
            self.endpoint,
            params=params,
        )
        self.content_type = response.headers["Content-Type"]
        assert isinstance(data, bytes)
        return data

    @raises(400, "Invalid request body or parameters")
    @raises(401, "Write privilege required for document '{document_id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Document '{document_id}' doesn't exists")
    @raises(
        409, "Specified revision {document_rev} is not the latest for target document"
    )
    async def _put(
        self, rev: str, data: bytes, content_type: str, **params: Any
    ) -> JsonDict:
        params["rev"] = rev
        _, json = await self._document._database._remote._put_bytes(
            self.endpoint,
            data=data,
            content_type=content_type,
            params=params,
        )
        self.content_type = content_type
        assert not isinstance(json, bytes)
        return json

    @raises(400, "Invalid request body or parameters")
    @raises(401, "Write privilege required for document '{document_id}'")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Specified database or document ID doesn't exists ({endpoint})")
    @raises(
        409, "Specified revision {document_rev} is not the latest for target document"
    )
    async def _delete(self, rev: str, **params: Any) -> JsonDict:
        params["rev"] = rev
        _, json = await self._document._database._remote._delete(
            self.endpoint,
            params=params,
        )
        self.content_type = None
        assert not isinstance(json, bytes)
        return json


class RemoteView:
    def __init__(self, database: "database.Database", ddoc: Optional[str], id: str):
        self._database = database
        self.ddoc = ddoc
        self.id = id

    @property
    def endpoint(self) -> str:
        assert self.ddoc is not None
        return (
            f"{self._database.endpoint}/_design/{_quote_id(self.ddoc)}/_view/"
            f"{_quote_id(self.id)}"
        )

    @raises(400, "Invalid request")
    @raises(401, "Read privileges required")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Specified database, design document or view is missing")
    async def _get(self, **params: Any) -> JsonDict:
        _, json = await self._database._remote._get(
            self.endpoint,
            params=params,
        )
        assert not isinstance(json, bytes)
        return json

    @raises(400, "Invalid request")
    @raises(401, "Write privileges required")
    @raises(403, "Access forbidden: {reason}")
    @raises(404, "Specified database, design document or view is missing")
    async def _post(self, keys: List[str], **params: Any) -> JsonDict:
        _, json = await self._database._remote._post(
            self.endpoint,
            data={"keys": keys},
            params=params,
        )
        assert not isinstance(json, bytes)
        return json



import asyncio

from aiocouch import CouchDB


# using the with statement, ensures a proper connection handling
async def main_with() -> None:
    # connect using username and password as credentials
    async with CouchDB(
        "http://localhost:5984", user="admin", password="admin"
    ) as couchdb:
        database = await couchdb["config"]

        async for doc in database.docs(["db-hta"]):
            print(doc)
    # connection is closed

    # connect using cookie
    async with CouchDB(
        "http://localhost:5984", cookie="ZGVtb0B0b2x0ZWNrLmNvbT..."
    ) as couchdb:
        await couchdb["_users"]
    # connection is closed


# storing the CouchDB instance allows for persistent connections
async def main_persistent() -> None:
    couchdb = CouchDB("http://localhost:5984", user="admin", password="admin")

    # optional credentials check
    await couchdb.check_credentials()

    # do something
    database = await couchdb["config"]
    async for doc in database.docs(["db-hta"]):
        print(doc)

    # close the connection without waiting
    asyncio.ensure_future(couchdb.close())

    return
    # connection is still open at this point, but will be closed soon after the return


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_with())
    loop.run_until_complete(main_persistent())



import pytest

from aiocouch import ConflictError, NotFoundError
from aiocouch.document import Document

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

text = (
    b"Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed non risus. Suspendisse "
    b"lectus tortor, dignissim sit amet, adipiscing nec, ultricies sed, dolor. Cras elementum "
    b"ultrices diam. Maecenas ligula massa, varius a, semper congue, euismod non, mi."
)

image = (
    b"RIFF\xb0\x00\x00\x00WEBPVP8 \xa4\x00\x00\x00P\x06\x00\x9d\x01*K\x002\x00?\xfd\xfe\xff"
    b"\x7f\xbf\xba\xb6\xb2>H\x03\xf0?\x89b\x00\xcc\x1c\x18\x05i\xc1\xdf\xfde\xe3\xfcr\xcb\xbb"
    b"DX^^\x9f\x9f\xcc/\xd2s]\x9a2j`\xeb\xaa@\x00\xd6\x8c\xfd;\xd5+\x9c\xb8\xb0\x9b\xabM\x07"
    b"\xc4\xd4\x07b\xcc\x1d\x13\x13\xa3\x9dR?\x9aJe\xe6\x1f\xbf\xed\x17\x19\xdc \x17\x92\xd9"
    b"\x8eR\x94\xad\x91+\xf1f@\x81\x9a\xcc{\x89\x84;\xd7\xdf\xffk\xea\xc0\xa8\xb9\xe1\xda\xea"
    b"\xe0\x07\x83\x12&\r\xae)$\x1b\xdc\xe9\xe5\xf0\xb0\xe0a\x0bP\xf1\x04iw\x03a\x86\rM\xe8"
    b"\xe9\xe0i\x08\xb1\xa0\x00\x00\x00"
)


async def test_save_text(doc: Document) -> None:
    await doc.save()
    att = doc.attachment("lipsum.txt")
    await att.save(text, "text/plain")


async def test_save_binary(doc: Document) -> None:
    await doc.save()
    att = doc.attachment("image.webp")
    await att.save(image, "image/webp")


async def test_save_on_unfetched(doc: Document) -> None:
    await doc.save()

    new_doc = Document(doc._database, doc["_id"])

    att = new_doc.attachment("lipsum.txt")
    with pytest.raises(ValueError):
        await att.save(text, "text/plain")

    new_doc = await doc._database.get(doc["_id"])
    att = new_doc.attachment("image.webp")
    await att.save(text, "text/plain")


async def test_get(doc: Document) -> None:
    await doc.save()
    await doc.attachment("image.webp").save(image, "image/webp")

    att = doc.attachment("image.webp")
    assert await att.exists()
    data = await att.fetch()
    assert data == image
    assert att.content_type == "image/webp"


async def test_get_on_unfetched(doc: Document) -> None:
    await doc.save()
    await doc.attachment("image.webp").save(image, "image/webp")

    new_doc = Document(doc._database, doc["_id"])
    att = new_doc.attachment("image.webp")
    data = await att.fetch()
    assert data == image


async def test_update(doc: Document) -> None:
    await doc.save()

    doc["value"] = 42
    await doc.save()

    att = doc.attachment("lipsum.txt")
    await att.save(text, "text/plain")

    att = doc.attachment("image.webp")
    await att.save(image, "image/webp")

    doc["value"] = 43
    await doc.save()

    new_doc = await doc._database.get(doc["_id"])
    assert new_doc["value"] == 43
    assert await new_doc.attachment("lipsum.txt").fetch() == text
    assert await new_doc.attachment("image.webp").fetch() == image


async def test_conflict(doc: Document) -> None:
    await doc.save()

    outdated = await doc._database.get(doc["_id"])

    doc["value"] = 42
    await doc.save()

    with pytest.raises(ConflictError):
        await outdated.attachment("image.webp").save(image, "image/webp")

    await doc.attachment("image.webp").save(image, "image/webp")


async def test_conflict_on_update(doc: Document) -> None:
    await doc.save()

    outdated = await doc._database.get(doc["_id"])

    att = doc.attachment("lipsum.txt")
    await att.save(text, "text/plain")

    with pytest.raises(ConflictError):
        await outdated.attachment("image.webp").save(image, "image/webp")

    await doc.attachment("image.webp").save(image, "image/webp")


async def test_delete(doc: Document) -> None:
    await doc.save()
    att = doc.attachment("lipsum.txt")
    await att.save(text, "text/plain")

    await att.delete()

    att = doc.attachment("lipsum.txt")
    with pytest.raises(NotFoundError):
        await att.fetch()

    assert not await att.exists()



# Copyright (c) 2019, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Any, Optional

from .document import Document
from .exception import ConflictError
from .remote import _quote_id
from .view import View


class DesignDocument(Document):
    _allowed_keys = [
        "language",
        "options",
        "filters",
        "lists",
        "rewrites",
        "shows",
        "updates",
        "validate_doc_update",
        "views",
    ]

    @property
    def endpoint(self) -> str:
        return f"{self._database.endpoint}/_design/{_quote_id(self.id)}"

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self._allowed_keys:
            super().__setitem__(key, value)
        else:
            raise KeyError(f"The key '{key}' is not allowed in an design document.")

    def view(self, view: str) -> View:
        return View(self._database, self.id, view)

    async def create_view(
        self,
        view: str,
        *,
        map_function: str,
        reduce_function: Optional[str] = None,
        exists_ok: bool = False,
    ) -> View:
        if "views" not in self:
            self["views"] = {}

        if view in self["views"] and not exists_ok:
            raise ConflictError(
                f"The view '{view}' does already exist in the design document {self.id}"
            )

        self["views"][view] = {"map": map_function}
        if reduce_function is not None:
            self["views"][view]["reduce"] = reduce_function
        self["language"] = "javascript"

        await self.save()

        return self.view(view)



from typing import Any, cast

import pytest

from aiocouch import ConflictError, NotFoundError
from aiocouch.database import Database
from aiocouch.document import Document

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_constructor(database: Database) -> None:
    from aiocouch.document import Document

    doc = Document(database, "foo", data={"foo": 42})

    assert doc.id == "foo"
    assert doc["_id"] == "foo"
    assert doc["foo"] == 42


async def test_constructor_with_wrong_type_for_data(
    database: Database, doc: Document
) -> None:
    with pytest.raises(TypeError):
        Document(database, "foo", data=cast(Any, 42))

    with pytest.raises(TypeError):
        Document(database, "foo", data=cast(Any, doc))


async def test_context_manager_by_creating_new_doc(database: Database) -> None:
    from aiocouch.document import Document

    new_doc_id = "new_doc"

    # Use context manager to create & save new doc with some data in it
    async with Document(database=database, id=new_doc_id) as document:
        assert (len(document.keys())) == 1
        assert document["_id"] == document.id == new_doc_id
        document["king"] = "elvis"

    # Verify whether the data was actually written to the server or not
    saved_doc = await database.get(new_doc_id)

    assert saved_doc["_id"] == saved_doc.id == new_doc_id
    assert saved_doc.rev is not None
    assert saved_doc.rev.startswith("1-")
    assert "king" in saved_doc
    assert saved_doc["king"] == "elvis"


async def test_context_manager_by_retrieving_existing_doc(
    filled_database: Database,
) -> None:
    """Test async context manager by retrieving an existing doc from
    filled_database fixture & verify that data was actually written to server"""

    from aiocouch.document import Document

    async with Document(database=filled_database, id="foo") as document:
        doc_keys = document.keys()
        assert len(doc_keys) == 4
        assert document["_id"] == document.id == "foo"
        assert document["_rev"] == document.rev
        assert "bar" in doc_keys
        assert document["bar"] is True
        assert document["bar2"] == 3


async def test_context_manager_with_data_parameter(database: Database) -> None:
    from aiocouch.document import Document

    new_doc_id = "new_doc"

    doc_data = {"king": "elvis"}

    # Create and save a new document with the data given to data parameter
    async with Document(database=database, id=new_doc_id, data=doc_data) as document:
        assert (len(document.keys())) == 2
        assert document.rev is None
        assert document["_id"] == document.id == new_doc_id
        assert "king" in document
        assert document["king"] == "elvis"

    # Verify that the data was actually written to server by context manager
    saved_doc = await database.get(new_doc_id)

    assert saved_doc["_id"] == saved_doc.id == new_doc_id
    assert saved_doc.rev is not None
    assert saved_doc.rev.startswith("1-")
    assert "king" in saved_doc
    assert saved_doc["king"] == "elvis"


async def test_save(database: Database) -> None:
    doc = await database.create("foo42")
    doc["bar"] = True
    await doc.save()

    keys = [key async for key in database.akeys()]

    assert "foo42" in keys
    assert len(keys) == 1


async def test_dict_from_doc(filled_database: Database) -> None:
    doc = await filled_database["foo"]

    doc_as_dict = doc.data
    assert isinstance(doc_as_dict, dict)

    doc = Document(filled_database, "buzz")
    doc_as_dict = doc.data
    assert doc_as_dict is None

    await doc.save()
    doc_as_dict = doc.data
    assert isinstance(doc_as_dict, dict)


async def test_fetch_clean_document(filled_database: Database) -> None:
    from aiocouch.document import Document

    doc = Document(filled_database, "foo")
    await doc.fetch()


async def test_fetch_dirty_document(database: Database) -> None:
    from aiocouch import ConflictError

    doc = await database.create("foo")

    doc["dirty"] = "shitz"
    with pytest.raises(ConflictError):
        await doc.fetch()


async def test_save_with_data(database: Database) -> None:
    doc = await database.create("foo")

    doc["blub"] = "blubber"

    await doc.save()

    assert doc["_id"] == doc.id == "foo"
    assert doc.rev is not None
    assert doc.rev.startswith("1-")
    assert doc._dirty_cache is False

    doc = await database["foo"]
    assert doc["_id"] == doc.id == "foo"
    assert doc.rev is not None
    assert doc.rev.startswith("1-")
    assert doc["blub"] == "blubber"
    assert doc._dirty_cache is False


async def test_save_fail_on_conflict(filled_database: Database) -> None:
    doc1 = await filled_database["foo"]
    doc2 = await filled_database["foo"]

    doc1["blub"] = "new"
    await doc1.save()

    doc2["blub"] = "bar"

    with pytest.raises(ConflictError):
        await doc2.save()

    conflicts = await doc1.conflicts()
    assert conflicts == []


async def test_save_fail_on_conflict_without_rev(database: Database) -> None:
    doc1 = await database.create("fou")
    doc2 = await database.create("fou")

    assert doc1.rev is None

    doc1["blub"] = "new"
    await doc1.save()

    assert doc1.rev is not None
    assert doc1.rev.startswith("1-")
    assert doc2.rev is None

    doc2["blub"] = "bar"

    with pytest.raises(ConflictError):
        await doc2.save()

    assert doc2.rev is None

    conflicts = await doc1.conflicts()
    assert conflicts == []


async def test_override_conflict(database: Database) -> None:
    doc1 = await database.create("fou")
    doc2 = await database.create("fou")

    doc1["blub"] = "new"
    await doc1.save()

    doc2["blub"] = "bar"

    try:
        await doc2.save()
        assert False
    except ConflictError:
        assert doc1.rev is not None
        doc2.rev = doc1.rev
        await doc2.save()

    conflicts = await doc2.conflicts()
    assert conflicts == []

    doc3 = await database["fou"]
    assert doc3.rev is not None
    assert doc3.rev.startswith("2-")
    assert doc3["blub"] == "bar"

    conflicts = await doc3.conflicts()
    assert conflicts == []


async def test_list_conflicts(database: Database) -> None:
    doc = await database.create("conflicted")

    doc["v"] = "1"
    await doc.save()
    assert doc.rev is not None
    assert doc.rev.startswith("1-")

    doc["v"] = "2"
    await doc.save()
    assert doc.rev.startswith("2-")

    doc.rev = "1-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    doc["v"] = "1a"
    await doc._put(doc._data, new_edits=False)

    conflicts = await doc.conflicts()
    assert conflicts == ["1-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"]

    doc.rev = "1-zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz"
    doc["v"] = "1z"
    await doc._put(doc._data, new_edits=False)

    conflicts = await doc.conflicts()
    assert conflicts == [
        "1-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
        "1-zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz",
    ]

    doc1a = await database.get("conflicted", rev="1-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
    await doc1a.delete()

    conflicts = await doc.conflicts()
    assert conflicts == [
        "1-zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz",
    ]

    await doc.fetch(discard_changes=True)
    assert doc.rev.startswith("2-")
    await doc.delete()

    doc = await database.get("conflicted")
    assert doc.rev == "1-zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz"
    assert doc["v"] == "1z"

    conflicts = await doc.conflicts()
    assert conflicts == []


async def test_list_revs(database: Database) -> None:
    doc = await database.create("revisions")

    assert doc.rev is None
    with pytest.raises(NotFoundError):
        await doc.revs()

    doc["v"] = "1"
    await doc.save()
    assert doc.rev is not None
    assert doc.rev.startswith("1-")
    revs = await doc.revs()
    assert revs == ["1-806db366bc1ba3ebef40f4b7c63a53af"]

    doc["v"] = "2"
    await doc.save()
    assert doc.rev.startswith("2-")
    revs = await doc.revs()
    assert revs == [
        "2-8ef50bd631ee32d350edd37cd66f2e31",
        "1-806db366bc1ba3ebef40f4b7c63a53af",
    ]


async def test_update(filled_database: Database) -> None:
    doc = await filled_database["foo"]

    assert doc.rev == doc["_rev"]
    assert doc.rev.startswith("1-")
    assert doc._dirty_cache is False

    del doc["bar"]
    doc["blub"] = "blubber"

    assert doc["_rev"] == doc.rev
    assert doc["_rev"].startswith("1-")
    assert doc._dirty_cache is True

    await doc.save()

    assert doc["_id"] == doc.id == "foo"
    assert doc["_rev"] == doc.rev
    assert doc["_rev"].startswith("2-")
    assert doc._dirty_cache is False

    doc = await filled_database["foo"]
    assert doc["_id"] == doc.id == "foo"
    assert doc["_rev"].startswith("2-")
    assert doc["_rev"] == doc.rev
    assert doc._dirty_cache is False
    assert doc["blub"] == "blubber"
    assert "bar" not in doc


async def test_delete(filled_database: Database) -> None:
    doc = await filled_database["foo"]

    await doc.delete()

    keys = [key async for key in filled_database.akeys()]

    assert len(keys) == 3
    assert sorted(keys) == ["baz", "baz2", "foo2"]


async def test_delete_dirty(filled_database: Database) -> None:
    from aiocouch import ConflictError

    doc = await filled_database["foo"]

    doc["fuzzy"] = "lizzy"

    with pytest.raises(ConflictError):
        await doc.delete()

    assert doc.rev
    assert doc.rev.startswith("1-")


async def test_safe_deleted(filled_database: Database) -> None:
    doc = await filled_database["foo"]

    await doc.delete()

    doc["Zebras"] = "are majestic"
    await doc.save()

    assert doc.rev
    assert doc.rev.startswith("3-")

    foo = await filled_database["foo"]
    assert foo.data is not None


async def test_copy(filled_database: Database) -> None:
    foo = await filled_database["foo"]
    assert foo.data is not None

    response = await foo.copy("foo_copy")

    assert response.status == 201
    assert "Location" in response.headers
    assert "Etag" in response.headers
    assert response.etag is not None
    assert response.etag.startswith("1-")

    foo_copy = await filled_database.get("foo_copy", rev=response.etag)

    assert response.status == 201

    assert foo_copy.data is not None
    assert foo_copy.data.keys() == foo.data.keys()
    for key in foo_copy.data.keys():
        if key == "_id":
            continue
        assert foo_copy[key] == foo[key]


async def test_get_info(database: Database) -> None:
    doc = await database.create("foo42")
    await doc.save()

    info = await doc.info()
    assert info["ok"]
    assert info["id"] == "foo42"
    assert info["rev"].startswith("1-")

    doc["bar"] = True
    await doc.save()

    info = await doc.info()
    assert info["ok"]
    assert info["id"] == "foo42"
    assert info["rev"].startswith("2-")

    await doc.delete()

    with pytest.raises(NotFoundError):
        await doc.info()


async def test_rev(doc: Document) -> None:
    assert doc.rev is None
    await doc.save()
    assert doc.rev is not None

    rev = doc.rev
    assert rev.startswith("1-")
    with pytest.raises(TypeError):
        doc.rev = 42
    assert doc.rev == rev


async def test_doc_update(doc: Document) -> None:
    assert "test" not in doc

    doc.update({"test": "value"})

    assert "test" in doc
    assert doc["test"] == "value"


async def test_doc_items_keys_values(doc: Document) -> None:
    assert list(doc.keys()) == ["_id"]
    assert list(doc.values()) == ["foo"]
    assert dict(doc.items()) == {"_id": "foo"}


async def test_filled_doc_items_keys_values(doc: Document) -> None:
    doc.update({"test": "value"})

    assert list(doc.keys()) == ["_id", "test"]
    assert list(doc.values()) == ["foo", "value"]
    assert dict(doc.items()) == {"_id": "foo", "test": "value"}


async def test_get(doc: Document) -> None:
    assert doc.get("foo") is None
    with pytest.raises(KeyError):
        doc["foo"]

    assert doc.get("foo", "baz") == "baz"
    with pytest.raises(KeyError):
        doc["foo"]

    assert doc.setdefault("foo") is None
    assert doc["foo"] is None

    assert doc.setdefault("baz", "bar") == "bar"
    assert doc["baz"] == "bar"

    assert doc.setdefault("baz", "kitty") == "bar"
    assert doc["baz"] == "bar"


async def test_repr(doc: Document) -> None:
    print(doc)


async def test_cache(doc: Document) -> None:
    assert doc._dirty_cache is True

    await doc.save()

    assert doc._dirty_cache is False

    doc["foo"] = {"hello": "kitty"}

    assert doc._dirty_cache is True

    await doc.save()

    assert doc._dirty_cache is False

    doc["foo"]["llama"] = "juicy"

    assert doc._dirty_cache is True

    await doc.save()

    assert doc._dirty_cache is False


async def test_fetch_rev(filled_database: Database) -> None:
    doc = await filled_database.get("foo")
    old_rev = doc.rev

    doc["Zebras"] = "are the best"
    await doc.save()

    assert doc.rev != old_rev

    await doc.fetch(rev=old_rev)

    assert doc.rev == old_rev
    assert "Zebras" not in doc

    doc2 = await filled_database.get("foo", rev=old_rev)
    assert doc.data == doc2.data


async def test_security_document_context_manager(database: Database) -> None:
    from aiocouch.document import SecurityDocument

    async with SecurityDocument(database=database) as sec_doc:
        assert sec_doc.members is None
        assert sec_doc.admins is None

        assert sec_doc.member_roles is None or sec_doc.member_roles == ["_admin"]
        assert sec_doc.admin_roles is None or sec_doc.admin_roles == ["_admin"]

        sec_doc.add_member("lennon")
        sec_doc.add_admin("elvis")

    async with SecurityDocument(database=database) as sec_doc:
        assert sec_doc.members is not None
        assert "lennon" in sec_doc.members

        assert sec_doc.admins is not None
        assert "elvis" in sec_doc.admins

        sec_doc.remove_member(member="lennon")
        sec_doc.remove_admin(admin="elvis")

    async with SecurityDocument(database=database) as sec_doc:
        assert sec_doc.members is not None
        assert "lennon" not in sec_doc.members

        assert sec_doc.admins is not None
        assert "elvis" not in sec_doc.admins


async def test_data(doc: Document) -> None:
    assert doc.data is None

    await doc.save()

    assert doc.data is not None

    assert "_id" in doc.data.keys()
    assert "_rev" in doc.data.keys()

    assert len(doc.data.keys()) == 2

    doc["zebra"] = "🦓"

    assert "zebra" in doc.data.keys()


async def test_json(doc: Document) -> None:
    assert doc.json == {}

    await doc.save()

    assert "_id" not in doc.json.keys()
    assert "_rev" not in doc.json.keys()

    doc["zebra"] = "🦓"

    assert doc.json != {}
    assert "zebra" in doc.json


async def test_clone_with_json(filled_database: Database) -> None:
    foo = await filled_database["foo"]

    clone = await filled_database.create("clone", data=foo.json)

    assert clone.id == "clone"
    assert clone.rev is None
    assert clone["bar"] is True
    assert clone["bar2"] == 3

    clone2 = Document(cast(Database, None), "clone2")
    clone2.update(foo.json)

    assert clone2.id == "clone2"
    assert clone2.rev is None
    assert clone2["bar"] is True
    assert clone2["bar2"] == 3

    clone3 = Document(cast(Database, None), "clone3", data=foo.json)

    assert clone3.id == "clone3"
    assert clone3.rev is None
    assert clone3["bar"] is True
    assert clone3["bar2"] == 3



import asyncio

from aiocouch import CouchDB


async def main() -> None:
    async with CouchDB(
        "http://localhost:5984", user="admin", password="admin"
    ) as couchdb:
        elapsed_time = 0

        while True:
            _, result = await couchdb._server._get("/_session")
            assert not isinstance(result, bytes)
            user = result["userCtx"]["name"]
            print(f"After {elapsed_time} sec: {user}")

            await asyncio.sleep(10)
            elapsed_time += 10


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



# Copyright (c) 2020, Adrien Vergé
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import cast

from .remote import RemoteAttachment


class Attachment(RemoteAttachment):
    """A local representation for the referenced CouchDB document attachment

    An instance of this class represents a local copy of an attachment of CouchDB
    documents.

    :ivar id: the id of the attachment
    :ivar content_type: the content type of the attachment, only available after
        :meth:`~aiocouch.attachment.Attachment.fetch()` has been called.

    :param `~aiocouch.document.Document` document: The correlated document
    :param id: the id of the attachment

    """

    async def exists(self) -> bool:
        """Checks if the attachment exists on the server

        :return: returns True if the attachment exists
        """
        return await self._exists()

    async def fetch(self) -> bytes:
        """Returns the content of the attachment

        :return: the attachment content

        """
        return await self._get()

    async def save(self, data: bytes, content_type: str) -> None:
        """Saves the given attachment content on the server

        :param data: the content of the attachment
        :param content_type: the content type of the given data. (See
            `Content type <https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.17>`_)

        """
        await self._put(await self._get_doc_rev(), data, content_type)
        # Parent document needs to have '_attachments' and '_rev' updated:
        await self._document.fetch()

    async def delete(self) -> None:
        """Deletes the attachment from the server"""
        await self._delete(await self._get_doc_rev())
        # Parent document needs to have '_attachments' and '_rev' updated
        await self._document.fetch()

    async def _get_doc_rev(self) -> str:
        if not self._document.exists:
            raise ValueError(
                "The document must be fetched or saved before updating attachments"
            )

        return cast(str, self._document["_rev"])
